@@ -1,17 +1,22 @@
 /**
  * @copyright
  * ====================================================================
- * Copyright (c) 2003 CollabNet.  All rights reserved.
+ *    Licensed to the Apache Software Foundation (ASF) under one
+ *    or more contributor license agreements.  See the NOTICE file
+ *    distributed with this work for additional information
+ *    regarding copyright ownership.  The ASF licenses this file
+ *    to you under the Apache License, Version 2.0 (the
+ *    "License"); you may not use this file except in compliance
+ *    with the License.  You may obtain a copy of the License at
  *
- * This software is licensed as described in the file COPYING, which
- * you should have received as part of this distribution.  The terms
- * are also available at http://subversion.tigris.org/license-1.html.
- * If newer versions of this license are posted there, you may use a
- * newer version instead, at your option.
+ *      http://www.apache.org/licenses/LICENSE-2.0
  *
- * This software consists of voluntary contributions made by many
- * individuals.  For exact contribution history, see the revision
- * history and logs, available at http://subversion.tigris.org/.
+ *    Unless required by applicable law or agreed to in writing,
+ *    software distributed under the License is distributed on an
+ *    "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+ *    KIND, either express or implied.  See the License for the
+ *    specific language governing permissions and limitations
+ *    under the License.
  * ====================================================================
  * @endcopyright
  *
@@ -20,8 +25,12 @@
  */
 
 #include "JNIUtil.h"
+#include "Array.h"
+
 #include <sstream>
+#include <vector>
 #include <locale.h>
+
 #include <apr_strings.h>
 #include <apr_tables.h>
 #include <apr_general.h>
@@ -55,7 +64,6 @@ apr_pool_t *JNIUtil::g_pool = NULL;
 std::list<SVNBase*> JNIUtil::g_finalizedObjects;
 JNIMutex *JNIUtil::g_finalizedObjectsMutex = NULL;
 JNIMutex *JNIUtil::g_logMutex = NULL;
-JNIMutex *JNIUtil::g_globalPoolMutext = NULL;
 bool JNIUtil::g_initException;
 bool JNIUtil::g_inInit;
 JNIEnv *JNIUtil::g_initEnv;
@@ -72,7 +80,7 @@ bool JNIUtil::JNIInit(JNIEnv *env)
   // Clear all standing exceptions.
   env->ExceptionClear();
 
-  // Remember the env paramater for the remainder of the request.
+  // Remember the env parameter for the remainder of the request.
   setEnv(env);
 
   // Lock the list of finalized objects.
@@ -118,39 +126,7 @@ bool JNIUtil::JNIGlobalInit(JNIEnv *env)
 
   apr_status_t status;
 
-  /* C programs default to the "C" locale. But because svn is supposed
-     to be i18n-aware, it should inherit the default locale of its
-     environment.  */
-  if (!setlocale(LC_ALL, ""))
-    {
-      if (stderr)
-        {
-          const char *env_vars[] = { "LC_ALL", "LC_CTYPE", "LANG", NULL };
-          const char **env_var = &env_vars[0], *env_val = NULL;
-          while (*env_var)
-            {
-              env_val = getenv(*env_var);
-              if (env_val && env_val[0])
-                break;
-              ++env_var;
-            }
 
-          if (!*env_var)
-            {
-              /* Unlikely. Can setlocale fail if no env vars are set? */
-              --env_var;
-              env_val = "not set";
-            }
-
-          fprintf(stderr,
-                  "%s: error: cannot set LC_ALL locale\n"
-                  "%s: error: environment variable %s is %s\n"
-                  "%s: error: please check that your locale name is "
-                  "correct\n",
-                  "svnjavahl", "svnjavahl", *env_var, env_val, "svnjavahl");
-        }
-      return FALSE;
-    }
 
   /* Initialize the APR subsystem, and register an atexit() function
    * to Uninitialize that subsystem at program exit. */
@@ -187,6 +163,19 @@ bool JNIUtil::JNIGlobalInit(JNIEnv *env)
       return FALSE;
     }
 
+  /* Create our top-level pool. */
+  g_pool = svn_pool_create(NULL);
+
+  apr_allocator_t* allocator = apr_pool_allocator_get(g_pool);
+
+  if (allocator)
+    {
+      /* Keep a maximum of 1 free block, to release memory back to the JVM
+         (and other modules). */
+      apr_allocator_max_free_set(allocator, 1);
+    }
+
+
 #ifdef ENABLE_NLS
 #ifdef WIN32
   {
@@ -198,7 +187,7 @@ bool JNIUtil::JNIGlobalInit(JNIEnv *env)
     apr_size_t inwords, outbytes;
     unsigned int outlength;
 
-    apr_pool_create(&pool, 0);
+    pool = svn_pool_create(g_pool);
     /* get dll name - our locale info will be in '../share/locale' */
     inwords = sizeof(ucs2_path) / sizeof(ucs2_path[0]);
     HINSTANCE moduleHandle = GetModuleHandle("libsvnjavahl-1");
@@ -217,24 +206,21 @@ bool JNIUtil::JNIGlobalInit(JNIEnv *env)
         return FALSE;
       }
     utf8_path[outlength - outbytes] = '\0';
-    internal_path = svn_path_internal_style(utf8_path, pool);
+    internal_path = svn_dirent_internal_style(utf8_path, pool);
     /* get base path name */
-    internal_path = svn_path_dirname(internal_path, pool);
-    internal_path = svn_path_join(internal_path, SVN_LOCALE_RELATIVE_PATH,
+    internal_path = svn_dirent_dirname(internal_path, pool);
+    internal_path = svn_dirent_join(internal_path, SVN_LOCALE_RELATIVE_PATH,
                                   pool);
     bindtextdomain(PACKAGE_NAME, internal_path);
-    apr_pool_destroy(pool);
+    svn_pool_destroy(pool);
   }
 #else
   bindtextdomain(PACKAGE_NAME, SVN_LOCALE_DIR);
 #endif
 #endif
 
-  /* Create our top-level pool. */
-  g_pool = svn_pool_create(NULL);
-
 #if defined(WIN32) || defined(__CYGWIN__)
-  /* See http://svn.collab.net/repos/svn/trunk/notes/asp-dot-net-hack.txt */
+  /* See http://svn.apache.org/repos/asf/subversion/trunk/notes/asp-dot-net-hack.txt */
   /* ### This code really only needs to be invoked by consumers of
      ### the libsvn_wc library, which basically means SVNClient. */
   if (getenv ("SVN_ASP_DOT_NET_HACK"))
@@ -263,10 +249,6 @@ bool JNIUtil::JNIGlobalInit(JNIEnv *env)
   if (isExceptionThrown())
     return false;
 
-  g_globalPoolMutext = new JNIMutex(g_pool);
-  if (isExceptionThrown())
-    return false;
-
   // initialized the thread local storage
   if (!JNIThreadData::initThreadData())
     return false;
@@ -289,15 +271,6 @@ apr_pool_t *JNIUtil::getPool()
   return g_pool;
 }
 
-/**
- * Return the mutex securing the global pool.
- * @return the mutex for the global pool
- */
-JNIMutex *JNIUtil::getGlobalPoolMutex()
-{
-  return g_globalPoolMutext;
-}
-
 void JNIUtil::raiseThrowable(const char *name, const char *message)
 {
   if (getLogLevel() >= errorLog)
@@ -333,6 +306,11 @@ JNIUtil::throwNativeException(const char *className, const char *msg,
   JNIEnv *env = getEnv();
   jclass clazz = env->FindClass(className);
 
+  // Create a local frame for our references
+  env->PushLocalFrame(LOCAL_FRAME_SIZE);
+  if (JNIUtil::isJavaExceptionThrown())
+    return;
+
   if (getLogLevel() >= exceptionLog)
     {
       JNICriticalSection cs(*g_logMutex);
@@ -345,45 +323,89 @@ JNIUtil::throwNativeException(const char *className, const char *msg,
       g_logStream << std::endl;
     }
   if (isJavaExceptionThrown())
-    return;
+    POP_AND_RETURN_NOTHING();
 
   jstring jmessage = makeJString(msg);
   if (isJavaExceptionThrown())
-    return;
+    POP_AND_RETURN_NOTHING();
   jstring jsource = makeJString(source);
   if (isJavaExceptionThrown())
-    return;
+    POP_AND_RETURN_NOTHING();
 
   jmethodID mid = env->GetMethodID(clazz, "<init>",
                                    "(Ljava/lang/String;Ljava/lang/String;I)V");
   if (isJavaExceptionThrown())
-    return;
+    POP_AND_RETURN_NOTHING();
   jobject nativeException = env->NewObject(clazz, mid, jmessage, jsource,
                                            static_cast<jint>(aprErr));
   if (isJavaExceptionThrown())
+    POP_AND_RETURN_NOTHING();
+
+  env->Throw(static_cast<jthrowable>(env->PopLocalFrame(nativeException)));
+}
+
+void
+JNIUtil::putErrorsInTrace(svn_error_t *err,
+                          std::vector<jobject> &stackTrace)
+{
+  if (!err)
     return;
 
-  env->DeleteLocalRef(clazz);
+  JNIEnv *env = getEnv();
+
+  // First, put all our child errors in the stack trace
+  putErrorsInTrace(err->child, stackTrace);
+
+  // Now, put our own error in the stack trace
+  jclass stClazz = env->FindClass("java/lang/StackTraceElement");
   if (isJavaExceptionThrown())
     return;
 
-  env->DeleteLocalRef(jmessage);
+  static jmethodID ctor_mid = 0;
+  if (ctor_mid == 0)
+    {
+      ctor_mid = env->GetMethodID(stClazz, "<init>",
+                                  "(Ljava/lang/String;Ljava/lang/String;"
+                                  "Ljava/lang/String;I)V");
+      if (isJavaExceptionThrown())
+        return;
+    }
+
+  jstring jdeclClass = makeJString("native");
   if (isJavaExceptionThrown())
     return;
 
-  env->DeleteLocalRef(jsource);
+  char *tmp_path;
+  char *path = svn_dirent_dirname(err->file, err->pool);
+  while (tmp_path = strchr(path, '/'))
+    *tmp_path = '.';
+
+  jstring jmethodName = makeJString(path);
   if (isJavaExceptionThrown())
     return;
 
-  env->Throw(static_cast<jthrowable>(nativeException));
+  jstring jfileName = makeJString(svn_dirent_basename(err->file, err->pool));
+  if (isJavaExceptionThrown())
+    return;
+
+  jobject jelement = env->NewObject(stClazz, ctor_mid, jdeclClass, jmethodName,
+                                    jfileName, (jint) err->line);
+
+  stackTrace.push_back(jelement);
+
+  env->DeleteLocalRef(stClazz);
+  env->DeleteLocalRef(jdeclClass);
+  env->DeleteLocalRef(jmethodName);
+  env->DeleteLocalRef(jfileName);
 }
 
 void JNIUtil::handleSVNError(svn_error_t *err)
 {
   std::string msg;
-  assembleErrorMessage(err, 0, APR_SUCCESS, msg);
+  assembleErrorMessage(svn_error_purge_tracing(err), 0, APR_SUCCESS, msg);
   const char *source = NULL;
 #ifdef SVN_DEBUG
+#ifndef SVN_ERR__TRACING
   if (err->file)
     {
       std::ostringstream buf;
@@ -393,8 +415,113 @@ void JNIUtil::handleSVNError(svn_error_t *err)
       source = buf.str().c_str();
     }
 #endif
-  throwNativeException(JAVA_PACKAGE "/ClientException", msg.c_str(),
-                       source, err->apr_err);
+#endif
+
+  // Much of the following is stolen from throwNativeException().  As much as
+  // we'd like to call that function, we need to do some manual stack
+  // unrolling, so it isn't feasible.
+
+  JNIEnv *env = getEnv();
+
+  // Create a local frame for our references
+  env->PushLocalFrame(LOCAL_FRAME_SIZE);
+  if (JNIUtil::isJavaExceptionThrown())
+    return;
+
+  jclass clazz = env->FindClass(JAVA_PACKAGE "/ClientException");
+  if (isJavaExceptionThrown())
+    POP_AND_RETURN_NOTHING();
+
+  if (getLogLevel() >= exceptionLog)
+    {
+      JNICriticalSection cs(*g_logMutex);
+      g_logStream << "Subversion JavaHL exception thrown, message:<";
+      g_logStream << msg << ">";
+      if (source)
+        g_logStream << " source:<" << source << ">";
+      if (err->apr_err != -1)
+        g_logStream << " apr-err:<" << err->apr_err << ">";
+      g_logStream << std::endl;
+    }
+  if (isJavaExceptionThrown())
+    POP_AND_RETURN_NOTHING();
+
+  jstring jmessage = makeJString(msg.c_str());
+  if (isJavaExceptionThrown())
+    POP_AND_RETURN_NOTHING();
+  jstring jsource = makeJString(source);
+  if (isJavaExceptionThrown())
+    POP_AND_RETURN_NOTHING();
+
+  jmethodID mid = env->GetMethodID(clazz, "<init>",
+                                   "(Ljava/lang/String;Ljava/lang/String;I)V");
+  if (isJavaExceptionThrown())
+    POP_AND_RETURN_NOTHING();
+  jobject nativeException = env->NewObject(clazz, mid, jmessage, jsource,
+                                           static_cast<jint>(err->apr_err));
+  if (isJavaExceptionThrown())
+    POP_AND_RETURN_NOTHING();
+
+#ifdef SVN_ERR__TRACING
+  // Add all the C error stack trace information to the Java Exception
+
+  // Get the standard stack trace, and vectorize it using the Array class.
+  static jmethodID mid_gst = 0;
+  if (mid_gst == 0)
+    {
+      mid_gst = env->GetMethodID(clazz, "getStackTrace",
+                                 "()[Ljava/lang/StackTraceElement;");
+      if (isJavaExceptionThrown())
+        POP_AND_RETURN_NOTHING();
+    }
+  Array stackTraceArray((jobjectArray) env->CallObjectMethod(nativeException,
+                                                             mid_gst));
+  std::vector<jobject> oldStackTrace = stackTraceArray.vector();
+
+  // Build the new stack trace elements from the chained errors.
+  std::vector<jobject> newStackTrace;
+  putErrorsInTrace(err, newStackTrace);
+
+  // Join the new elements with the old ones
+  for (std::vector<jobject>::const_iterator it = oldStackTrace.begin();
+            it < oldStackTrace.end(); ++it)
+    {
+      newStackTrace.push_back(*it);
+    }
+
+  jclass stClazz = env->FindClass("java/lang/StackTraceElement");
+  if (isJavaExceptionThrown())
+    POP_AND_RETURN_NOTHING();
+
+  jobjectArray jStackTrace = env->NewObjectArray(newStackTrace.size(), stClazz,
+                                                 NULL);
+  if (isJavaExceptionThrown())
+    POP_AND_RETURN_NOTHING();
+
+  int i = 0;
+  for (std::vector<jobject>::const_iterator it = newStackTrace.begin();
+       it < newStackTrace.end(); ++it)
+    {
+      env->SetObjectArrayElement(jStackTrace, i, *it);
+      ++i;
+    }
+
+  // And put the entire trace back into the exception
+  static jmethodID mid_sst = 0;
+  if (mid_sst == 0)
+    {
+      mid_sst = env->GetMethodID(clazz, "setStackTrace",
+                                 "([Ljava/lang/StackTraceElement;)V");
+      if (isJavaExceptionThrown())
+        POP_AND_RETURN_NOTHING();
+    }
+  env->CallVoidMethod(nativeException, mid_sst, jStackTrace);
+  if (isJavaExceptionThrown())
+    POP_AND_RETURN_NOTHING();
+#endif
+
+  env->Throw(static_cast<jthrowable>(env->PopLocalFrame(nativeException)));
+
   svn_error_clear(err);
 }
 
@@ -492,7 +619,7 @@ bool JNIUtil::isJavaExceptionThrown()
 }
 
 const char *
-JNIUtil::thrownExceptionToCString()
+JNIUtil::thrownExceptionToCString(SVN::Pool &in_pool)
 {
   const char *msg;
   JNIEnv *env = getEnv();
@@ -509,7 +636,7 @@ JNIUtil::thrownExceptionToCString()
         }
       jstring jmsg = (jstring) env->CallObjectMethod(t, getMessage);
       JNIStringHolder tmp(jmsg);
-      msg = tmp.pstrdup(getRequestPool()->pool());
+      msg = tmp.pstrdup(in_pool.getPool());
       // ### Conditionally add t.printStackTrace() to msg?
     }
   else
@@ -640,31 +767,10 @@ jobject JNIUtil::createDate(apr_time_t time)
     return NULL;
 
   env->DeleteLocalRef(clazz);
-  if (isJavaExceptionThrown())
-    return NULL;
 
   return ret;
 }
 
-/**
- * Return the request pool. The request pool will be destroyed after each
- * request (call).
- * @return the pool to be used for this request
- */
-Pool *JNIUtil::getRequestPool()
-{
-  return JNIThreadData::getThreadData()->m_requestPool;
-}
-
-/**
- * Set the request pool in thread local storage.
- * @param pool  the request pool
- */
-void JNIUtil::setRequestPool(Pool *pool)
-{
-  JNIThreadData::getThreadData()->m_requestPool = pool;
-}
-
 /**
  * Create a Java byte array from an array of characters.
  * @param data      the character array
@@ -672,16 +778,18 @@ void JNIUtil::setRequestPool(Pool *pool)
  */
 jbyteArray JNIUtil::makeJByteArray(const signed char *data, int length)
 {
-  if (data == NULL || length == 0)
-    // a NULL or empty will create no Java array
-    return NULL;
+  if (data == NULL)
+    {
+      // a NULL will create no Java array
+      return NULL;
+    }
 
   JNIEnv *env = getEnv();
 
   // Allocate the Java array.
   jbyteArray ret = env->NewByteArray(length);
-  if (isJavaExceptionThrown())
-    return NULL;
+  if (isJavaExceptionThrown() || ret == NULL)
+      return NULL;
 
   // Access the bytes.
   jbyte *retdata = env->GetByteArrayElements(ret, NULL);
@@ -785,6 +893,9 @@ svn_error_t *JNIUtil::preprocessPath(const char *&path, apr_pool_t *pool)
         return svn_error_createf(SVN_ERR_BAD_URL, NULL,
                                  _("URL '%s' contains a '..' element"),
                                  path);
+
+      /* strip any trailing '/' */
+      path = svn_uri_canonicalize(path, pool);
     }
   else  /* not a url, so treat as a path */
     {
@@ -806,10 +917,11 @@ svn_error_t *JNIUtil::preprocessPath(const char *&path, apr_pool_t *pool)
          the performance penalty.
        */
 
-      path = svn_path_internal_style(path, pool);
+      path = svn_dirent_internal_style(path, pool);
+
+      /* For kicks and giggles, let's absolutize it. */
+      SVN_ERR(svn_dirent_get_absolute(&path, path, pool));
     }
-    /* strip any trailing '/' */
-    path = svn_path_canonicalize(path, pool);
 
   return NULL;
 }
