@@ -2,20 +2,26 @@
  * reporter.c : `reporter' vtable routines for updates.
  *
  * ====================================================================
- * Copyright (c) 2000-2006, 2008 CollabNet.  All rights reserved.
+ *    Licensed to the Apache Software Foundation (ASF) under one
+ *    or more contributor license agreements.  See the NOTICE file
+ *    distributed with this work for additional information
+ *    regarding copyright ownership.  The ASF licenses this file
+ *    to you under the Apache License, Version 2.0 (the
+ *    "License"); you may not use this file except in compliance
+ *    with the License.  You may obtain a copy of the License at
  *
- * This software is licensed as described in the file COPYING, which
- * you should have received as part of this distribution.  The terms
- * are also available at http://subversion.tigris.org/license-1.html.
- * If newer versions of this license are posted there, you may use a
- * newer version instead, at your option.
+ *      http://www.apache.org/licenses/LICENSE-2.0
  *
- * This software consists of voluntary contributions made by many
- * individuals.  For exact contribution history, see the revision
- * history and logs, available at http://subversion.tigris.org/.
+ *    Unless required by applicable law or agreed to in writing,
+ *    software distributed under the License is distributed on an
+ *    "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+ *    KIND, either express or implied.  See the License for the
+ *    specific language governing permissions and limitations
+ *    under the License.
  * ====================================================================
  */
 
+#include "svn_dirent_uri.h"
 #include "svn_path.h"
 #include "svn_types.h"
 #include "svn_error.h"
@@ -26,6 +32,8 @@
 #include "svn_props.h"
 #include "repos.h"
 #include "svn_private_config.h"
+#include "private/svn_dep_compat.h"
+#include "private/svn_fspath.h"
 
 #define NUM_CACHED_SOURCE_ROOTS 4
 
@@ -77,14 +85,25 @@ typedef struct path_info_t
   apr_pool_t *pool;            /* Container pool */
 } path_info_t;
 
+/* Describes the standard revision properties that are relevant for
+   reports.  Since a particular revision will often show up more than
+   once in the report, we cache these properties for the time of the
+   report generation. */
+typedef struct revision_info_t
+{
+  svn_revnum_t rev;            /* revision number */
+  svn_string_t* date;          /* revision timestamp */
+  svn_string_t* author;        /* name of the revisions' author */
+} revision_info_t;
+
 /* A structure used by the routines within the `reporter' vtable,
    driven by the client as it describes its working copy revisions. */
 typedef struct report_baton_t
 {
   /* Parameters remembered from svn_repos_begin_report2 */
   svn_repos_t *repos;
-  const char *fs_base;         /* FS path corresponding to wc anchor */
-  const char *s_operand;       /* Anchor-relative wc target (may be empty) */
+  const char *fs_base;         /* fspath corresponding to wc anchor */
+  const char *s_operand;       /* anchor-relative wc target (may be empty) */
   svn_revnum_t t_rev;          /* Revnum which the edit will bring the wc to */
   const char *t_path;          /* FS path the edit will bring the wc to */
   svn_boolean_t text_deltas;   /* Whether to report text deltas */
@@ -111,6 +130,11 @@ typedef struct report_baton_t
   path_info_t *lookahead;
   svn_fs_root_t *t_root;
   svn_fs_root_t *s_roots[NUM_CACHED_SOURCE_ROOTS];
+
+  /* Cache for revision properties. This is used to eliminate redundant
+     revprop fetching. */
+  apr_hash_t* revision_infos;
+
   apr_pool_t *pool;
 } report_baton_t;
 
@@ -153,6 +177,7 @@ static svn_error_t *
 read_string(const char **str, apr_file_t *temp, apr_pool_t *pool)
 {
   apr_uint64_t len;
+  apr_size_t size;
   char *buf;
 
   SVN_ERR(read_number(&len, temp, pool));
@@ -161,7 +186,7 @@ read_string(const char **str, apr_file_t *temp, apr_pool_t *pool)
      len + 1 wraps around and we end up passing 0 to apr_palloc(),
      thus getting a pointer to no storage?  Probably not (16 exabyte
      string, anyone?) but let's be future-proof anyway. */
-  if (len + 1 < len)
+  if (len + 1 < len || len + 1 > APR_SIZE_MAX)
     {
       /* xgettext doesn't expand preprocessor definitions, so we must
          pass translatable string to apr_psprintf() function to create
@@ -174,8 +199,9 @@ read_string(const char **str, apr_file_t *temp, apr_pool_t *pool)
                                len);
     }
 
-  buf = apr_palloc(pool, len + 1);
-  SVN_ERR(svn_io_file_read_full(temp, buf, len, NULL, pool));
+  size = (apr_size_t)len;
+  buf = apr_palloc(pool, size+1);
+  SVN_ERR(svn_io_file_read_full2(temp, buf, size, NULL, NULL, pool));
   buf[len] = 0;
   *str = buf;
   return SVN_NO_ERROR;
@@ -419,6 +445,53 @@ change_file_prop(report_baton_t *b, void *file_baton, const char *name,
   return b->editor->change_file_prop(file_baton, name, value, pool);
 }
 
+/* For the report B, return the relevant revprop data of revision REV in
+   REVISION_INFO. The revision info will be allocated in b->pool.
+   Temporaries get allocated on SCRATCH_POOL. */
+static  svn_error_t *
+get_revision_info(report_baton_t *b,
+                  svn_revnum_t rev,
+                  revision_info_t** revision_info,
+                  apr_pool_t *scratch_pool)
+{
+  apr_hash_t *r_props;
+  svn_string_t *cdate, *author;
+  revision_info_t* info;
+
+  /* Try to find the info in the report's cache */
+  info = apr_hash_get(b->revision_infos, &rev, sizeof(rev));
+  if (!info)
+    {
+      /* Info is not available, yet.
+         Get all revprops. */
+      SVN_ERR(svn_fs_revision_proplist(&r_props,
+                                       b->repos->fs,
+                                       rev,
+                                       scratch_pool));
+
+      /* Extract the committed-date. */
+      cdate = apr_hash_get(r_props, SVN_PROP_REVISION_DATE,
+                           APR_HASH_KEY_STRING);
+
+      /* Extract the last-author. */
+      author = apr_hash_get(r_props, SVN_PROP_REVISION_AUTHOR,
+                            APR_HASH_KEY_STRING);
+
+      /* Create a result object */
+      info = apr_palloc(b->pool, sizeof(*info));
+      info->rev = rev;
+      info->date = cdate ? svn_string_dup(cdate, b->pool) : NULL;
+      info->author = author ? svn_string_dup(author, b->pool) : NULL;
+
+      /* Cache it */
+      apr_hash_set(b->revision_infos, &info->rev, sizeof(rev), info);
+    }
+
+  *revision_info = info;
+  return SVN_NO_ERROR;
+}
+
+
 /* Generate the appropriate property editing calls to turn the
    properties of S_REV/S_PATH into those of B->t_root/T_PATH.  If
    S_PATH is NULL, this is an add, so assume the target starts with no
@@ -431,12 +504,13 @@ delta_proplists(report_baton_t *b, svn_revnum_t s_rev, const char *s_path,
                 void *object, apr_pool_t *pool)
 {
   svn_fs_root_t *s_root;
-  apr_hash_t *s_props, *t_props, *r_props;
+  apr_hash_t *s_props, *t_props;
   apr_array_header_t *prop_diffs;
   int i;
   svn_revnum_t crev;
   const char *uuid;
-  svn_string_t *cr_str, *cdate, *last_author;
+  svn_string_t *cr_str;
+  revision_info_t* revision_info;
   svn_boolean_t changed;
   const svn_prop_t *pc;
   svn_lock_t *lock;
@@ -450,21 +524,17 @@ delta_proplists(report_baton_t *b, svn_revnum_t s_rev, const char *s_path,
       SVN_ERR(change_fn(b, object,
                         SVN_PROP_ENTRY_COMMITTED_REV, cr_str, pool));
 
-      SVN_ERR(svn_fs_revision_proplist(&r_props, b->repos->fs, crev, pool));
+      SVN_ERR(get_revision_info(b, crev, &revision_info, pool));
 
       /* Transmit the committed-date. */
-      cdate = apr_hash_get(r_props, SVN_PROP_REVISION_DATE,
-                           APR_HASH_KEY_STRING);
-      if (cdate || s_path)
+      if (revision_info->date || s_path)
         SVN_ERR(change_fn(b, object, SVN_PROP_ENTRY_COMMITTED_DATE,
-                          cdate, pool));
+                          revision_info->date, pool));
 
       /* Transmit the last-author. */
-      last_author = apr_hash_get(r_props, SVN_PROP_REVISION_AUTHOR,
-                                 APR_HASH_KEY_STRING);
-      if (last_author || s_path)
+      if (revision_info->author || s_path)
         SVN_ERR(change_fn(b, object, SVN_PROP_ENTRY_LAST_AUTHOR,
-                          last_author, pool));
+                          revision_info->author, pool));
 
       /* Transmit the UUID. */
       SVN_ERR(svn_fs_get_uuid(b->repos->fs, &uuid, pool));
@@ -599,7 +669,10 @@ fake_dirent(const svn_fs_dirent_t **entry, svn_fs_root_t *root,
   else
     {
       ent = apr_palloc(pool, sizeof(**entry));
-      ent->name = svn_path_basename(path, pool);
+      /* ### All callers should be updated to pass just one of these
+             formats */
+      ent->name = (*path == '/') ? svn_fspath__basename(path, pool)
+                                 : svn_relpath_basename(path, pool);
       SVN_ERR(svn_fs_node_id(&ent->id, root, path, pool));
       ent->kind = kind;
       *entry = ent;
@@ -668,7 +741,7 @@ add_file_smartly(report_baton_t *b,
          starting with '/', so make sure o_path always starts with a '/'
          too. */
       if (*o_path != '/')
-        o_path = apr_pstrcat(pool, "/", o_path, NULL);
+        o_path = apr_pstrcat(pool, "/", o_path, (char *)NULL);
 
       SVN_ERR(svn_fs_closest_copy(&closest_copy_root, &closest_copy_path,
                                   b->t_root, o_path, pool));
@@ -748,7 +821,6 @@ update_entry(report_baton_t *b, svn_revnum_t s_rev, const char *s_path,
   void *new_baton;
   svn_checksum_t *checksum;
   const char *hex_digest;
-  int distance;
 
   /* For non-switch operations, follow link_path in the target. */
   if (info && info->link_path && !b->is_switch)
@@ -786,7 +858,7 @@ update_entry(report_baton_t *b, svn_revnum_t s_rev, const char *s_path,
   related = FALSE;
   if (s_entry && t_entry && s_entry->kind == t_entry->kind)
     {
-      distance = svn_fs_compare_ids(s_entry->id, t_entry->id);
+      int distance = svn_fs_compare_ids(s_entry->id, t_entry->id);
       if (distance == 0 && !any_path_info(b, e_path)
           && (requested_depth <= wc_depth || t_entry->kind == svn_node_file))
         {
@@ -806,8 +878,7 @@ update_entry(report_baton_t *b, svn_revnum_t s_rev, const char *s_path,
             }
         }
 
-      if (distance != -1 || b->ignore_ancestry)
-        related = TRUE;
+      related = (distance != -1 || b->ignore_ancestry);
     }
 
   /* If there's a source and it's not related to the target, nuke it. */
@@ -956,8 +1027,6 @@ delta_dirs(report_baton_t *b, svn_revnum_t s_rev, const char *s_path,
   apr_hash_t *s_entries = NULL, *t_entries;
   apr_hash_index_t *hi;
   apr_pool_t *subpool;
-  const svn_fs_dirent_t *s_entry, *t_entry;
-  void *val;
   const char *name, *s_fullpath, *t_fullpath, *e_fullpath;
   path_info_t *info;
 
@@ -984,6 +1053,8 @@ delta_dirs(report_baton_t *b, svn_revnum_t s_rev, const char *s_path,
 
       while (1)
         {
+          const svn_fs_dirent_t *s_entry, *t_entry;
+
           svn_pool_clear(subpool);
           SVN_ERR(fetch_path_info(b, &name, &info, e_path, subpool));
           if (!name)
@@ -1005,10 +1076,10 @@ delta_dirs(report_baton_t *b, svn_revnum_t s_rev, const char *s_path,
               continue;
             }
 
-          e_fullpath = svn_path_join(e_path, name, subpool);
-          t_fullpath = svn_path_join(t_path, name, subpool);
+          e_fullpath = svn_relpath_join(e_path, name, subpool);
+          t_fullpath = svn_fspath__join(t_path, name, subpool);
           t_entry = apr_hash_get(t_entries, name, APR_HASH_KEY_STRING);
-          s_fullpath = s_path ? svn_path_join(s_path, name, subpool) : NULL;
+          s_fullpath = s_path ? svn_fspath__join(s_path, name, subpool) : NULL;
           s_entry = s_entries ?
             apr_hash_get(s_entries, name, APR_HASH_KEY_STRING) : NULL;
 
@@ -1052,9 +1123,10 @@ delta_dirs(report_baton_t *b, svn_revnum_t s_rev, const char *s_path,
                hi;
                hi = apr_hash_next(hi))
             {
+              const svn_fs_dirent_t *s_entry;
+
               svn_pool_clear(subpool);
-              apr_hash_this(hi, NULL, NULL, &val);
-              s_entry = val;
+              s_entry = svn__apr_hash_index_val(hi);
 
               if (apr_hash_get(t_entries, s_entry->name,
                                APR_HASH_KEY_STRING) == NULL)
@@ -1071,11 +1143,11 @@ delta_dirs(report_baton_t *b, svn_revnum_t s_rev, const char *s_path,
                     continue;
 
                   /* There is no corresponding target entry, so delete. */
-                  e_fullpath = svn_path_join(e_path, s_entry->name, subpool);
+                  e_fullpath = svn_relpath_join(e_path, s_entry->name, subpool);
                   SVN_ERR(svn_repos_deleted_rev(svn_fs_root_fs(b->t_root),
-                                                svn_path_join(t_path,
-                                                              s_entry->name,
-                                                              subpool),
+                                                svn_fspath__join(t_path,
+                                                                 s_entry->name,
+                                                                 subpool),
                                                 s_rev, b->t_rev,
                                                 &deleted_rev, subpool));
 
@@ -1089,9 +1161,10 @@ delta_dirs(report_baton_t *b, svn_revnum_t s_rev, const char *s_path,
       /* Loop over the dirents in the target. */
       for (hi = apr_hash_first(pool, t_entries); hi; hi = apr_hash_next(hi))
         {
+          const svn_fs_dirent_t *s_entry, *t_entry;
+
           svn_pool_clear(subpool);
-          apr_hash_this(hi, NULL, NULL, &val);
-          t_entry = val;
+          t_entry = svn__apr_hash_index_val(hi);
 
           if (is_depth_upgrade(wc_depth, requested_depth, t_entry->kind))
             {
@@ -1118,12 +1191,12 @@ delta_dirs(report_baton_t *b, svn_revnum_t s_rev, const char *s_path,
                   apr_hash_get(s_entries, t_entry->name, APR_HASH_KEY_STRING)
                   : NULL;
               s_fullpath = s_entry ?
-                  svn_path_join(s_path, t_entry->name, subpool) : NULL;
+                  svn_fspath__join(s_path, t_entry->name, subpool) : NULL;
             }
 
           /* Compose the report, editor, and target paths for this entry. */
-          e_fullpath = svn_path_join(e_path, t_entry->name, subpool);
-          t_fullpath = svn_path_join(t_path, t_entry->name, subpool);
+          e_fullpath = svn_relpath_join(e_path, t_entry->name, subpool);
+          t_fullpath = svn_fspath__join(t_path, t_entry->name, subpool);
 
           SVN_ERR(update_entry(b, s_rev, s_fullpath, s_entry, t_fullpath,
                                t_entry, dir_baton, e_fullpath, NULL,
@@ -1151,7 +1224,7 @@ drive(report_baton_t *b, svn_revnum_t s_rev, path_info_t *info,
 
   /* Compute the target path corresponding to the working copy anchor,
      and check its authorization. */
-  t_anchor = *b->s_operand ? svn_path_dirname(b->t_path, pool) : b->t_path;
+  t_anchor = *b->s_operand ? svn_fspath__dirname(b->t_path, pool) : b->t_path;
   SVN_ERR(check_auth(b, &allowed, t_anchor, pool));
   if (!allowed)
     return svn_error_create
@@ -1159,7 +1232,7 @@ drive(report_baton_t *b, svn_revnum_t s_rev, path_info_t *info,
        _("Not authorized to open root of edit operation"));
 
   /* Collect information about the source and target nodes. */
-  s_fullpath = svn_path_join(b->fs_base, b->s_operand, pool);
+  s_fullpath = svn_fspath__join(b->fs_base, b->s_operand, pool);
   SVN_ERR(get_source_root(b, &s_root, s_rev));
   SVN_ERR(fake_dirent(&s_entry, s_root, s_fullpath, pool));
   SVN_ERR(fake_dirent(&t_entry, b->t_root, b->t_path, pool));
@@ -1258,9 +1331,12 @@ finish_report(report_baton_t *b, apr_pool_t *pool)
   {
     svn_error_t *err = drive(b, s_rev, info, pool);
     if (err == SVN_NO_ERROR)
-      return b->editor->close_edit(b->edit_baton, pool);
-    svn_error_clear(b->editor->abort_edit(b->edit_baton, pool));
-    return err;
+      return svn_error_trace(b->editor->close_edit(b->edit_baton, pool));
+
+    return svn_error_trace(
+                svn_error_compose_create(err,
+                                         b->editor->abort_edit(b->edit_baton,
+                                                               pool)));
   }
 }
 
@@ -1278,7 +1354,7 @@ write_path_info(report_baton_t *b, const char *path, const char *lpath,
 
   /* Munge the path to be anchor-relative, so that we can use edit paths
      as report paths. */
-  path = svn_path_join(b->s_operand, path, pool);
+  path = svn_relpath_join(b->s_operand, path, pool);
 
   lrep = lpath ? apr_psprintf(pool, "+%" APR_SIZE_T_FMT ":%s",
                               strlen(lpath), lpath) : "-";
@@ -1348,9 +1424,8 @@ svn_repos_finish_report(void *baton, apr_pool_t *pool)
 
   finish_err = finish_report(b, pool);
   close_err = svn_io_file_close(b->tempfile, pool);
-  if (finish_err)
-    svn_error_clear(close_err);
-  return finish_err ? finish_err : close_err;
+
+  return svn_error_trace(svn_error_compose_create(finish_err, close_err));
 }
 
 svn_error_t *
@@ -1358,7 +1433,7 @@ svn_repos_abort_report(void *baton, apr_pool_t *pool)
 {
   report_baton_t *b = baton;
 
-  return svn_io_file_close(b->tempfile, pool);
+  return svn_error_trace(svn_io_file_close(b->tempfile, pool));
 }
 
 /* --- BEGINNING THE REPORT --- */
@@ -1391,11 +1466,11 @@ svn_repos_begin_report2(void **report_baton,
      keep track of them. */
   b = apr_palloc(pool, sizeof(*b));
   b->repos = repos;
-  b->fs_base = apr_pstrdup(pool, fs_base);
+  b->fs_base = svn_fspath__canonicalize(fs_base, pool);
   b->s_operand = apr_pstrdup(pool, s_operand);
   b->t_rev = revnum;
-  b->t_path = switch_path ? switch_path
-    : svn_path_join(fs_base, s_operand, pool);
+  b->t_path = switch_path ? svn_fspath__canonicalize(switch_path, pool)
+                          : svn_fspath__join(b->fs_base, s_operand, pool);
   b->text_deltas = text_deltas;
   b->requested_depth = depth;
   b->ignore_ancestry = ignore_ancestry;
@@ -1405,6 +1480,8 @@ svn_repos_begin_report2(void **report_baton,
   b->edit_baton = edit_baton;
   b->authz_read_func = authz_read_func;
   b->authz_read_baton = authz_read_baton;
+  b->revision_infos = apr_hash_make(pool);
+  b->pool = pool;
 
   SVN_ERR(svn_io_open_unique_file3(&b->tempfile, NULL, NULL,
                                    svn_io_file_del_on_pool_cleanup,
